'use strict';
var superagent = require('superagent');
var Immutable = require('immutable');
var _ = require('underscore');
var LegacyI3APIHeaders = require('legacy/I3APIHeaders');
var Router = require('Router');
var Wish = require('Wish');
function errorFromSuperagentErrorAndRes(error, res) {
    if (error) {
        return {
            message: error.message,
            line: error.line,
            stack: error.stack,
            name: error.name
        };
    } else if (res && !res.ok) {
        return {
            status: res.status,
            body: res.body
        };
    }
}
function Retrieve(path, method, url, mapper, options) {
    mapper = mapper || (x => x);
    return {
        name: 'API/Retrieve',
        transaction: ws => ws.setIn(path, Wish.Pending()),
        effect: world => API.request(world, url, _.partial(RetrievalWasResolved, path, mapper), _.partial(RetrievalWasRejected, path), options)
    };
}
function RetrievalWasResolved(path, mapper, value) {
    return {
        name: 'API/RetrievalWasResolved',
        transaction: ws => ws.setIn(path, Wish.Available(mapper(value)))
    };
}
function RetrievalWasRejected(path, error) {
    return {
        name: 'API/RetrievalWasRejected',
        transaction: ws => ws.setIn(path, Wish.Error(error))
    };
}
function forwardedHeaders(world) {
    var headers = world.getIn([
        'ServerOnly',
        'Request',
        'headers'
    ]);
    return headers ? headers.toJS() : {};
}
var API = {
    request: function (world, route, resolvingActionCreator, rejectingActionCreator, options) {
        options = options || {};
        var accept = options.accept || 'application/json';
        var method = options.method || 'GET';
        var payload = options.payload;
        var type = options.type || 'json';
        var extraHeaders = options.extraHeaders || {};
        var url = Router.urlForRoute(route);
        console.log('API.%s: %s', method, url);
        new superagent.Request(method, url).set(forwardedHeaders(world)).set('Accept', accept).set(extraHeaders).type(type).send(payload).end(function (error, res) {
            error = errorFromSuperagentErrorAndRes(error, res);
            if (error) {
                console.error('API.%s ERROR: %s', method, JSON.stringify(error));
                world.performAction(rejectingActionCreator(Immutable.fromJS(error)));
            } else {
                world.performAction(resolvingActionCreator(Immutable.fromJS(res.body), Immutable.fromJS(res.headers)));
            }
        });
    },
    get: function (world, path, route, mapper) {
        return Wish.performOnce(world, path, Retrieve(path, 'GET', route, mapper));
    },
    legacyGet: function (world, path, route, mapper) {
        return Wish.performOnce(world, path, Retrieve(path, 'GET', route, mapper, { extraHeaders: LegacyI3APIHeaders }));
    },
    constants: function (world, ...constants) {
        var params = {};
        constants.forEach(constant => {
            if (!world.getIn([
                    'Constants',
                    constant
                ])) {
                world.performAction({
                    name: 'API/ConstantsWillFetch',
                    transaction: ws => ws.setIn([
                        'Constants',
                        constant
                    ], Wish.Pending())
                });
                params[constant] = true;
            }
        });
        if (Object.keys(params).length === 0) {
            return;
        }
        return API.request(world, Router.route('api/v3/constants', params), function resolve(body, headers) {
            return {
                name: 'API/ConstantsResolved',
                transaction: ws => constants.reduce((ws, constant) => {
                    return ws.setIn([
                        'Constants',
                        constant
                    ], Wish.Available(body.getIn([
                        'data',
                        constant
                    ])));
                }, ws)
            };
        }, function reject(error) {
            return {
                name: 'API/ConstantsRejected',
                transaction: ws => constants.reduce((ws, constant) => {
                    return ws.setIn([
                        'Constants',
                        constant
                    ], Wish.Error(error));
                }, ws)
            };
        });
    }
};
module.exports = API;
